<!-- TODO: move in mytheme -->
<style>
    .reveal p {
	    text-align: left;
	}
</style>


# Ordonnancement

--

## Ordonancement préemptif

Le noyaux se charge de distribuer les processus sur les différents CPU:
- l’ordonnanceur attribue un CPU à un processus, généralement **pour une tranche de temps précise appelée quantum**;
- l’ordonnanceur choisi quel est le nouveau processus qui va être alloué à ce CPU une fois le quantum ou le processus terminé;


<div class="text-block">

Toutefois un processus peut libérer un CPU volontairement avant la fin du quantum si:
- il se met en attente d’une ressource;
- il reçoit un signal de suspension (c.f. SIGSTOP, SIGSTP);
- il appel la fonction `sched_yield` (c.f. man).

</div>

--

## Odonancement sous Linux

Implémentation:
- il existe une liste de processus pour différentes priorités statiques [0-99];
- les processus de hautes priorités sont toujours exécutés d’abord.

![](img/sched-priority.svg) <!-- .element: width="180%"-->

--

## Priorité statique 0

Trois modes d’ordonnancement sont disponibles en priorité statique 0:

- **standard (i.e. par défaut – SCHED_OTHER)**;
- pour processus à lourde charge de calcul (SCHED_BATCH);
- pour processus à très très faible priorité (SCHED_IDLE).

--

## Priorité statique 0 - SCHED_OTHER


Cette stratégie permet de s’assurer que chaque processus sera traité après avoir eu un certain nombre de déni de CPU qui dépends de sa priorité et de celle des autres processus.

- le processus de la liste est choisi par rapport à une **priorité dynamique**;
- priorité dynamique = *valeur nice* + nombre de quantum en état prêt sans avoir de processeur à disposition;
- La *valeur nice* d’un processus est attribuée par la commande `nice` ou par la fonction C `setpriority`.

--

## Ordonnancement "TR" - FIFO

Implémentation:
- les processus sont rangés dans une liste lors de leur soumission;
- un nouveau processus est placé en queue de liste;
- chaque processus est exécuté sans interruption (i.e. non préemptif, sauf cas de mise en attente).

![](img/sched-fifo.svg) <!-- .element: width="120%"-->

--

## Odonnancement "TR" - Round Robin

Implémentation:
- les processus sont placés en queue de liste lors de leur soumission (idem FIFO);
- chaque processus est exécuté uniquement pour un quantum de temps (préemptif) puis est replacé en fin de liste.

![](img/sched-RR.svg) <!-- .element: width="120%"-->

--

## Contrôle d'ordonnancement

L’ordonnancement d’un processus peut être contrôlé par les fonctions et structures suivantes pour la priorité statique:

```c
#include <sched.h>
int sched_setscheduler(pid_t pid, int policy, const struct sched_param *param);
int sched_getscheduler(pid_t pid);

struct sched_param {
	...
	int sched_priority; /* prioritée statique */
	...
};
```
Policy peut prendre les valeurs: SCHED_OTHER,SCHED_BATCH, SCHED_IDLE,SCHED_FIFO, SCHED_RR.

La priorité dynamique peut être contrôlé en utilisant:

```c
int setpriority(int which, int who, int prio); /* priorité dynamique, nice */
int getpriority(int which, int who);
```

--

## Commutation de contexte

Lorsque qu’un processus doit en remplacer un autre sur un CPU une commutation de contexte a lieu:
- suspension de l’exécution du processus et sauvegarde de son contexte;
- rétablissement de l’état du CPU à l’état sauvegardé lors de la suspension du processus qui reprend son exécution;
- mise en exécution du nouveau processus.

Au passage il faut:
- Écrire sur le disque les pages modifiées (pages sales);
- Mettre à jour les informations du noyau pour tenir compte du changement de processus actif (par exemple re-calcul de la priorité des processus).

N.B.: Un changement de contexte est couteuse d’où l’avantage des threads.

--

## Le processus dans tous ses états

<p style="text-align:center;"><img src="img/proc-states.svg" width=55% alt="Process states"></p>

| ps | Constante noyau      | Description de l’état |
| -- | -------------------- | --------------------- |
| R  | TASK_RUNNING         | En exécution ou prêt à être exécuté (i.e. dans une file d’attente) |
| S  | TASK_INTERRUPTIBLE   | En attente d’une ressource (Sleep) et interruptible par un signal |
| D  | TASK_UNINTERRUPTIBLE | En attente d’une ressource mais ne peut pas être interrompu |
| T  | TASK_STOPPED         | Suspendu (sToppé) |
| Z  | EXIT_ZOMBIE          | Zombie en attente d’un wait de la part du parent. |
| X  | EXIT_DEAD            | Terminé, mort, ne devrais jamais être observé. |
<!-- .element: style="font-size: 0.7em;" -->