#include "file_transmission.h"
#include <stdio.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

void prepare_address( struct sockaddr_in *address, const char *host, int port ) {
  size_t addrSize = sizeof( address );
  memset(address, 0, addrSize);
  address->sin_family = AF_INET;
  inet_pton( AF_INET, (char*) address, &(address->sin_addr) );
  address->sin_port = htons(port);
}

int makeSocket( const char *host, int port ) {
  struct sockaddr_in address;
  int sock = socket(AF_INET, SOCK_STREAM, 0);
  if( sock < 0 ) {
    die("Failed to create socket");
  }
  prepare_address( &address, host, port );
  if( connect(sock, (struct sockaddr *) &address, sizeof(address)) < 0) {
    die("Failed to connect with server");
  }
  return sock;
}


int main(int argc, char *argv[]) {
  int sock,file;
  char *host;
  int port;
  char *filename;
  int filenameLength;

  if (argc != 4) {
    fprintf(stderr, "USAGE: %s <host> <port> <filename>\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  host = argv[1];
  port = atoi(argv[2]);
  filename = argv[3];
  filenameLength = strlen( filename );

  /* Open the file */
  if( (file = open(filename,O_WRONLY|O_CREAT|O_TRUNC,0600)) < 0 ) {
    die("Failed to create the local file");
  }

  sock = makeSocket( host, port );

  /* Send the filename */
  if( write(sock,filename,filenameLength) != filenameLength ) {
    die( "Cannot send the filename to retrieve" );
  }

  /* Receive the file */
  if( copy( sock, file ) < 0 ) {
    die( "Cannot receive the file" );
  }

  close(sock);
  close(file);


  exit(EXIT_SUCCESS);
}
