<!-- TODO: move in mytheme -->
<style>
    .reveal p {
	    text-align: left;
	}
    .reveal ul {
        display: block;
    }
    .reveal code {
        font-size: 0.9em;
    }
    .reveal pre code { <!-- is there a way to say in the previous one "except if in pre" -->
        font-size: 1em;
    }
</style>


# Drivers et Modules

Guillaume Chanel <!-- .element: style="text-align:center" -->

---

# Définitions et généralités

--

## Périphérique (Device)

- Un périphérique ou "device" est un matériel connecté à une unité centrale.

*Exemples de périphériques: une souris, un terminal (clavier + écran), un disque dur, un lecteur DVD,  un capteur de température, un affichage LED, ...*

- Sous linux les périphériques sont souvent associés à des fichiers de périphériques ou "device files" qui permettent de communiquer avec le matériel.

*Exemples de device files: /dev/sda (disque dure), /dev/video0 (webcam), /dev/ttyS0 (port série), ...*

--

## Périphérique virtuel (Virtual device)


Un périphérique virtuel ou "virtual device" est un fichier de **périphérique qui n'est pas associé à un hardware**. Cela permet par exemple de faire de l'émulation de périphérique.

*Exemples de périphériques virtuels:*
- */dev/null émule un trou noir à octets,*
- */dev/zero génère des zéros,*
- */dev/urandom génère des nombres aléatoires,*
- *les consoles virtuelles (/dev/tty0, /dev/tty1, ...),*
- *les pseudo-terminaux (/dev/pts/0, /dev/pts/1, ...),*
- *des fichiers émulant des disques dures ou "loop devices" (/dev/loop1, /dev/loop2, ...)*

Nous implémenterons un périphérique virtuel en TP.

--

## Types de périphériques

Les prériphériques peuvent être de deux types:

- **Character devices:**

Le noyau communique avec ces périphériques octet par octet: on peu lire et écrire le nombre d'octets que l'on souhaite.

Ce sont les périphériques les plus communs (ports séries, terminaux, souris, webcam ...)

- **Block devices:**

On ne peut lire des données que par blocs (souvent des blocs de 4K).

C'est par exemple le cas des disques dures ou des CD-ROM qui nécéssitent de transiter des quantités d'information importante rapidement.

--

## Drivers et modules

Un **module** est un code compilé qui est ajouté au noyau alors que ce dernier est en exécution (drivers, protocoles réseaux, systèmes de fichiers, transformations cryptographiques, etc.)

Chaque périphérique est géré par une couche logicielle appelée **driver** (ou pilote). On distingue les drivers:

- intégrés au noyau: il s'agit de drivers qui sont compilés en même temps que le noyau (e.g. terminaux);

- implémentés sous forme de module, ce qui corresponds à la pluspart des drivers aujourd'hui.

Linux étant un **système monolithique**, les drivers sont exécutés dans l'espace noyau. Il existe des systèmes d'exploitations de type **micro-moyau** ou les drivers sont exécutés dans l'espace utilisateur (c.f. minix).

--

## Drivers

Le but d'un driver est de gérer les I/O d'un périphérique. Il devra donc être capable de répondre aux demandes des programmes de l'espace utilisateur. Pour cela il doit pouvoir répondre aux appels systèmes suivant:
- `open`: initialiser un nouveau canal de communication permettant d'accéder au périphérique;
- `read` / `write`: lire des données ou envoyer des données au périphérique;
- `close`: fermer le canal de communication vers le périphérique uniquement si il n'y a plus de descripteur de fichier référençant ce canal.
- `lseek`: ce déplacer dans le périphérique si cela est possible (rarement pour  un périphérique physique),
- etc.

--

## Gérer les modules

<!-- .slide: style="font-size: 0.8em" -->

La commande modprobe permet de gérer les modules (voir aussi les fichiers de configuration /etc/modules-load.d).

- insérer un module dans le noyau (un module peut avoir des paramètres)
```bash
# si le module est présent dans le dossier /lib/modules/linux-version. Insère aussi les dépendances.
$ modprobe module [param1=value1 param2=value2]
# si l'on veut insérer un module dans un autre dossier (sans ses dépendances)
$ insmod /path/to/my/module.ko [param1=value1 ...]
```

- retirer un module du noyau
```bash
$ modeprobe -r module
$ rmmod module # commande alternative
```

- lister les module chargés par le noyau
```bash
$ lsmod
```

- obtenir des informations sur un module
```bash
$ modinfo module
```

--

## Identification des périphériques

Le noyau identifie chaque périphérique grâce à une paire de nombres (`majeur`, `mineur`):

- **le nombre majeur**: indique généralement le driver qui permettra d'exploiter le périphérique;
- **le nombre mineur**: indique à quel périphérique du driver l'on fait référence.

<pre><code class="bash" data-noescape data-trim>
$ ls /dev/
brw-rw----   1 root   disk        <strong>8,   0</strong> Mar 25 15:03 sda
brw-rw----   1 root   disk        <strong>8,   1</strong> Mar 25 15:03 sda1
brw-rw----   1 root   disk        <strong>8,   2</strong> Mar 25 15:03 sda2
brw-rw----+  1 root   optical    <strong>11,   0</strong> Mar 25 15:03 sr0  # DVD géré par un autre driver
</code></pre>

notes:
on peu penser au nombre mineur comme un index utiliser par le driver pour trouver le bon prériphérique dans une table de périphériques. C'est d'ailleur un possibilité d'implémenter cette table sois même dans le noyau

Il est possible d'avoir plusieurs drivers utilisant le même numéro majeur mais c'est rarement le cas

--

## Identification des périphériques (2)

En C c'est le type `dev_t`, accompagné de macros, qui permet de gérer ces identifiants:

```c
#include <linux/types.h>
#include <linux/kdev_t.h>
dev_t dev = MKDEV(int major, int minor) // créer des identifiants de device
MAJOR(dev_t dev) // obtenir le majeur d'un device
MINOR(dev_t dev) // obtenir le mineur d'un device
```

--

## Créer un device file

La commande mknod permet de créer un fichier de périphérique associé aux idenfiants (`majeur`, `mineur`)

```bash
mknod [-m mode] filename type major minor
```

- `filename`: le nom du fichier à créer (e.g. /dev/mydevice)
- `type`: le type de périphérique (i.e. 'b' pour block, 'c' pour character)
- `major`, `minor`: les identifiants du périphérique


---

# Environement de travail

--

## Les sources du noyau

Les modules sont **liés aux sources du noyau**. Il est donc nécessaire d'avoir les sources pour pouvoir compiler un module. On peut:
- télécharger les sources et les compiler ([https://www.kernel.org/](https://www.kernel.org/));
- utiliser le gestionnaire de package pour télécharger les sources (générallement dans /usr/lib/modules/linux-version)

--

## Makefiles

La compilation de modules fait référence à des makefile du noyau. Le makefile d'un module à donc un format particulier:
```makefile
LINUX_VER=$(shell uname -r)

# votre code est dans my-module.c
# indique au noyau que un fichier objet dois être compilé pour ce module
obj-m := my-module.o
# voir module-objs pour des modules à plusieurs fichiers

build:
    # Fait l'hypothèse que les sources du noyau sont dans /lib/modules/...
    # -C execute le makefile du noyau dans le dossier spécifié
    # M=path indique au noyau le dossier du module (pour pouvoir obtenir les obj-m)
    # puis le noyau execute la target modules qui compile les objets dans obj-m
    make -C /lib/modules/$(LINUX_VER)/build M=$(PWD) modules
```

--

## Pas de glib  !!! &#128561;

Comme les modules résident dans le noyau, aucune fonction de la glib n'est disponible...

- pas de `malloc`, `calloc`, ...
- pas de `printf`
- ...

--

## [Fonctions disponibles](https://www.kernel.org/doc/htmldocs/kernel-api/) &#128524;

Heureusement les programmeurs du noyau ont pensé à nous:

- `printk` existe
- `kmalloc` existe

En fait beaucoup d'autres fonctions sont (re)-implémentées comme `strncpy`. Toutefois leur comportement peut varié de celui de la norme ANSI.

Une [documentation](https://www.kernel.org/doc/htmldocs/kernel-api/) de l'API du noyaux linux est disponible.

--

## `printk` - messages de log

Très utilisé pour le debuggage, `printk` permet d'envoyer des messages au système de log du noyau suivant un format similaire à `printf`:
```c
printk(KERN_ERR "Error (%d): %s", number, string);
```

`KERN_ERR` est le niveau de log et peut prendre d'autre valeurs (`KERN_EMERG`, `KERN_ALERT`, `KERN_CRIT`, `KERN_ERR`, `KERN_WARNING`, `KERN_INFO`, `KERN_DEBUG`, ...)


Ces messages:
- apparaissent sur la console¹ si le niveau de log est suffisant (paramètre du noyau);
- sont visualisables par la commande `dmesg`.

¹: pas sur les pseudo-terminaux fenêtrés <!-- .element: style="font-size:0.8em;" -->

--

## kmalloc

```c
#include <linux/slab.h>
void * kmalloc(size_t size, gfp_t flags)
void kfree(void *ptr)
```

- `size`: taille en octet de la mémoire à allouer;
- `flags`: le type de mémoire à allouer;
- retourne un pointeur `ptr` sur la mémoire allouée, le libérer avec `kfree`.

Dans la majorité des cas le type de mémoire utilisé est `GFP_KERNEL`, cet appel implique que le processus courrant peut-être mis en attente de la ressouce mémoire (voir `GFP_NOWAIT`, `GFP_ATOMIC` pour des alternatives).

---

# Module: vue générale

![Graphique des structures importantes du noyau](img/struct_overview.svg)

---

# Déclaration et configuration d'un périphérique

--

## Information sur le module

Tout module va inclure les fichiers d'entête suivants:
```c
#include <linux/module.h>
#include <linux/init.h>
```

De plus des macros permettent de donner des informations sur le module:
```c
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("Guillaume Chanel <Guillaume.Chanel@unige.ch>");
MODULE_DESCRIPTION("This is a wonderful module");
MODULE_VERSION("1.0");
// Voir également MODULE_ALIAS, MODULE_DEVICE_TABLE
```

La macro `MODULE_LICENSE` est particulièrement importante et peut prendre plusieurs valeurs. Sans cette macro le module est considéré comme non-opensource et on dit que le noyau est "contaminé" (tainted).

Note: Par convention on les retrouve générallement en fin de module.

--

## Initialisation et terminaison du module

Notre module devra implémenter deux fonctions qui seront appellées lorsque le module et chargé et retiré du noyau:
- l'utilisation de `modprobe` ou `insmod` appellera une fonction d'initialisation;
- l'utilisation de `modprobe -r` ou `rmmod` appellera une fonction de terminaison.

Il ne faut pas confondre ces fonctions avec l'ouverture (`open`) et la fermeture (`close / release`) d'un périphérique.

--

## Initialisation du module

<!-- .slide: style="font-size: 0.95em" -->

Une fonction doit être dédiée à l'initialisation du module:

```c
static int __init my_wonderfull_init(void) {
    ...
}
module_init(my_wonderfull_init)
```
Cette fonction doit préparer les périphériques, ce qui veux dire:
1. réserver des identifiants pour les périphériques (majeur / mineurs), on dit **inscrire des périphériques**;
2. initialiser les données nécéssaires au bon fonctionnement du périphérique;
3. ajouter le périphérique au système d'exploitation.

**Attention, lorsqu'un périphérique est ajouté au système il peut alors être utiliser immédiatement** (même si la fonction d'initialisation n'est pas terminée). Il faut donc que le prériphérique sois prêt à fonctionner avant de l'ajouter au noyau.

note: ici je parle uniquement de périhpérique mais le noyau peu en fait enregister plein d'autre équipements (facility): périphériques bloques, systèmes de fichiers, transformations cryptographiques, etc.

--

## Initialisation du module

Une fonction doit être dédiée à l'initialisation du module:

```c
static int __init my_wonderfull_init(void) {
    ...
}
module_init(my_wonderfull_init)
```

En cas d'erreur dans la fonction d'initialisation il faut:
- **penser à nettoyer ce qui a été fait (voir terminaison du module)**
- retourner le code d'erreur donné par la fonction fautive.

--

## Initialisation du module

Une fonction doit être dédiée à l'initialisation du module:

```c
static int __init my_wonderfull_init(void) {
    ...
}
module_init(my_wonderfull_init)
```

Explication du code:
- `__init` est un mot clef du noyau qui s'assure que la fonction sera libérée après la phase d'initialisation. Cela permet d'économiser l'espace mémoire du noyau;
- `module_init()` est une macro du noyau qui permets de lui indiquer où se trouve la fonction d'initialisation.

--

## Terminaison du module

Une fonction doit être dédiée à la terminaison du module:

```c
static void __exit my_wonderfull_finale(void) {
    ...
}
module_exit(my_wonderfull_finale)
```

Cette fonction doit:

<ol reversed>
  <li>retirer les périphériques déjà ajoutés au système;</li>
  <li>eventuellement libérer des données allouées;
  <li>libérer les identifiants (majeur / mineurs), on dit désinscrire des périphériques.</li>
</ol>

**En cas d'erreur lors de l'initialisation** il faut aussi effectuer ces trois étapes pour tout périphérique enregistré.

--

## Terminaison du module

Une fonction doit être dédiée à la terminaison du module:

```c
static void __exit my_wonderfull_finale(void) {
    ...
}
module_exit(my_wonderfull_finale)
```

Explication du code:
- `__exit` place la fonction dans une section spécifique du noyau et ne peut être utiliser QUE lorsque le module est retiré du noyau (`__exit` n'est pas obligatoire);
- `module_exit()` est une macro du noyau qui permets de lui indiquer où se trouve la fonction de terminaison.

Note: __exit peut être retiré si l'on souhaite appeller la fonction depuis l'initialization pour désallouer les prériphériques précédement alloué en cas d'erreur.

De plus, dans le cas ou le module n'est pas dechargeable / retirable, le mot clef __exit indique au noyau qu'il peu supprimer la fonction.

--

## 1. Inscrire des périphériques

<!-- .slide: style="font-size:0.95em" -->

Pour inscrire un périphérique caractère on réserve un ensemble de :
```c
#include <linux/fs.h>
// Pour obtenir un numéro majeur en spécifiant le premier mineur et le nombre de périphériques
int alloc_chrdev_region(dev_t *dev, unsigned int firstminor, unsigned int count, char *name);
// Pour spécifier les numéros majeur, mineur et le nombre de périphériques
int register_chrdev_region(dev_t dev, unsigned int count, char *name);
```

- `dev`: numéro de device retourné par la fonction (majeur + premier mineur). On peut aussi le spécifier avec `register_chrdev_region`;
- `firstminor`: premier numéro mineur référencant notre premier périphérique (généralement 0);
- `count`: nombre de périphériques que l'on souhaite inscrire (déterminera le plus haut nombre mineur de nos prériphériques);
- `name`: nom de notre module / ensemble de périphériques;
- retourne 0 en cas de succès, un code d'erreur sinon.

Rappel: on peu gérer le type `dev_t` avec des macros:
<!-- .element: style="margin-bottom:0" -->

```c
MKDEV(MAJOR(dev), MINOR(dev) + i) // référencer notre ième périphérique
```
<!-- .element: style="margin-top:0" -->


--

## 1. Inscrire des périphériques

Une fois inscrit, on peut retrouver le numéro majeur de nos périphériques dans /proc/devices:
```bash
$ cat /proc/devices
Character devices:
...
137 mydevice
...
```

--

## 1. Desinscrire des périphériques

Pour desinscrire un périphérique lors de la terminaison d'un module ou d'une erreure on utilise:
```c
void unregister_chrdev_region(dev_t first, unsigned int count);
```
- `first`: identifiant du premier périphérique;
- `count`: nombre de périphériques.

--

## 2. Définition d'un périphérique

<!-- .slide: style="font-size:0.95em" -->

En général, on définie une structure globale qui va contenir les données globales des nos périphériques.

```c
struct mydevice {
    char *string_data; /* ou tout autre données que l'on souhaite */
    int int_data; /* ou tout autre données que l'on souhaite */
    struct cdev mycdev;
}
```
Cette structure contiens une structure `cdev` qui représente notre périphérique caractère dans le noyau.

```c
struct cdev {
    ...
    struct file_operations ops;
    ...
}
```

**C'est la stucture `file_operations` qui va nous permettre de définir les actions a réaliser lorsque nous recevons des appels système.**

--

## 2. Initialiser un périphérique

Avant d'ajouter le périphérique il faut l'initialiser:
```c
#include <linux/cdev.h>
void cdev_init(struct cdev *cdev, struct file_operations *fops);
```

- `cdev`: tel que définie précédement;
- `fops`: doit être initialisée correctement (voir section suivante "répondre aux appels systèmes").

--

## 3. Ajouter un périphérique

Pour **ajouter un périphérique** on utilise:
```c
int cdev_add(struct cdev *dev, dev_t num, unsigned int count);
```
- `dev`: structure initialisée avec `cdev_init`;
- `num`: le numéro périphérique;
- `count`: générallement 1.
- retourne 0 en case de succès, un code d'erreur sinon.

N.B.: il existe des cas où un prériphérique correspond à plusieurs numéro de prériphériques. C'est pourquoi `count` existe.


Pour **supprimer un périphérique**:
```c
void cdev_del(struct cdev *dev);
```

--

## Paramètres des modules

La commande modprobe peut passer des paramètres de configuration à notre module:
```bash
$ modprobe module [param1=value1 param2=value2]
```

Dans notre module ces paramètres correspondent à des variables globales qui sont rendues disponibles par des macros:
```c
static int param1 = 0;
static char *param2 = "default value";
module_param(param1, int, S_IRUGO);
module_param(param2, charp, S_IRUGO);
```

Note: on peut aussi utiliser des tableaux avec `module_param_array`

--

## Paramètres des modules

```c
module_param(name, type, perms);
```

- `name`: le nom de la variable;
- `type`: le type de la variable; à choisir parmis `bool`, `charp`, `int`, `long`, `short`, `uint`, `ulong`, `ushort`;
- `perms`: chaque paramètre du module apparait comme entrée dans le système de fichier spécial `sysfs`. Les droits de ce fichier sont définis pas le paramètre `perms`:
    - S_IRUGO - Read for User / Group / Others;
    - S_IWUSR - Write for User root);
    - ... (les valeur habituelles fonctionnent)

---

# Répondre aux appels systèmes

--

## `file`: canal de communication

La structure `file` représente un **canal de communication** / un fichier ouvert (peut-être un fichier de périphérique):

```c
#include <linux/fs.h>
struct file {
    mode_t f_mode;
    loff_t f_pos;
    unsigned int f_flags;
    struct file_operations *f_op;
    void *private_data;
    struct dentry *f_dentry;
}
```

Cette structure est crée lors d'un appel `open` et supprimée lorsque tous les descripteurs de fichiers ont été fermé (dernier `close`).

Il s'agit donc d'une **entrée de la table des canaux**, comme nous l'avons vu précédement.

--

## `file`: canal de communication

Quelques champs importants:

```c [4,6-7]
#include <linux/fs.h>
struct file {
    mode_t f_mode;
    loff_t f_pos;
    unsigned int f_flags;
    struct file_operations *f_op;
    void *private_data;
    struct dentry *f_dentry;
}
```

- `f_pos`: la position actuelle dans le fichier;
- `f_op`: un pointeur vers la liste des fonctions permettant de manipuler le fichier (lire, ecrire, etc.);
- `private_data`: un pointeur pour stocker des données dépendant du type de canal (e.g. un pointeur sur notre device `struct mydevice`).

--

## `file_operations`

C'est la structure `file_operations` qui determine comment notre module réagit aux appels systèmes. Cette struture est généralement déclarée de manière globale dans le module.

```c
#include <linux/module.h>
#include <linux/fs.h>
struct file_operations my_fops = {
    .owner = THIS_MODULE,
    .open = my_open,
    .release = my_release,
    .read = my_read,
    .write = my_write,
    .llseek = my_lseek,
    ...
};
```
La macro `THIS_MODULE` retourne un pointeur sur notre module.

--

## `file_operations`

```c
struct file_operations my_fops = {
    .owner = THIS_MODULE,
    .open = my_open,
    .release = my_release,
    .read = my_read,
    .write = my_write,
    .llseek = my_lseek,
    ...
};
```

Chaque élément de la structure est un pointeur sur une fonction de notre module (sauf `owner`):
```c
int my_open(struct inode *i, struct file *filp) {
    // Actions à réaliser pour ouvrir notre périphérique
}


ssize_t my_read(struct file *filp, char __user *buffer, size_t, loff_t *offset) {
    // Actions à réaliser pour lire notre périphérique
}

...
```

--

## `open`

```c
int (*open)(struct inode *inode, struct file *filp);
```

`open` est appellée à chaque ouverture du périphérique. Son but est de préparer le driver à traiter les opérations futures:
- retourner une erreure si le driver n'est pas prêt;
- créer les données privées `filp->private_data`  (si nécéssaire).

On peut mettre l'entrée `open` de la structure `f_ops` à `NULL`. Dans ce cas le noyau crée simplement la structure `file` et notre driver n'est pas notifié d'une ouverture.

Examples de préparation de device:
- initialiser un buffer de communication;
- envoyer une commande d'initalisation au hardware pour le premier `open`.

--

## `release`

```c
int (*release) (struct inode *inode, struct file *filp);
```

`release` est appellée uniquement lorsque il n'y a plus de descripteurs de fichiers pointants sur `filp` (un `release` pour un `open`).

Toute mémoire initialisée par `open` doit être libérée par `release`.

On peut mettre l'entrée `release` de la structure `f_ops` à `NULL`. Dans ce cas le noyau libère simplement la structure `file` et notre driver n'est pas notifié d'une fermeture.


--

## `read`

```c
ssize_t (*read) (struct file *filp, char __user *buff, size_t len, loff_t *f_pos);
```

`read` est appelé à chaque lecture du périphérique (voir l'appel système correspondant).

- `filp`: le fichier concerné par la lecture;
- `buff`, `len`: le buffer à remplir avec les données lues et la quantité de données demandées. Si on le souhaite **on peu lire moins de données que demandées**.
- `f_pos`:  position courante dans le fichier. **Il faut mettre à jour cette position en fonction du nombre de bytes effectivement lus.**
- doit retourner la quantité de bytes effectivement lus.

--

## `read`

<!-- .slide: style="font-size: 0.9em" -->

```c
ssize_t (*read) (struct file *filp, char __user *buff, size_t len, loff_t *f_pos);
```

**Attention**: comme indiqué par la macro `__user` **le pointeur `buff` est une adresse de l'espace mémoire du processus; il faut donc convertir cette adresse dans l'espace d'adressage du noyau.**

Pour cela on utilise:
```c
#include <asm/uaccess.h>
unsigned long copy_to_user(void __user *to, const void *from, unsigned long count);
```

- `to`: l'adresse du processus vers laquelle copier les données;
- `from`: les données de l'espace noyau à copiées;
- `count`: la quantité de bytes à copier.
- retourne la quantié de bytes qu'il reste à copier, si elle est différente de 0 alors un problème à été rencontré.

--

## `write`

```c
ssize_t (*write) (struct file *filp, const char __user *buff, size_t len, loff_t *f_pos);
```

`write` est appelé à chaque écriture du périphérique (voir l'appel système correspondant).

- `filp`: le fichier concerné par l'écriture;
- `buff`, `len`: le buffer et la quantité de données à écrire. Si on le souhaite **on peu écrire moins de données que demandées**.
- `f_pos`:  position courante dans le fichier. **Il faut mettre à jour cette position en fonction du nombre de bytes effectivement écrits.**
- doit retourner la quantité de bytes effectivement écrits.

--


## `write`

<!-- .slide: style="font-size: 0.9em" -->

```c
ssize_t (*write) (struct file *filp, const char __user *buff, size_t len, loff_t *f_pos);
```

**Attention**: comme indiqué par la macro `__user` **le pointeur `buff` est une adresse de l'espace mémoire du processus; il faut donc convertir cette adresse dans l'espace d'adressage du noyau.**

Pour cela on utilise:
```c
#include <asm/uaccess.h>
unsigned long copy_from_user(void *to, const void __user *from, unsigned long count);
```

- `to`: l'adresse noyau vers laquelle copier les données;
- `from`: les données à copiées depuis l'espace du processus;
- `count`: la quantité de bytes à copier.
- retourne la quantié de bytes qu'il reste à copier, si elle est différente de 0 alors un problème à été rencontré.

--

## `lseek`

```c
loff_t (*llseek) (struct file *filp, loff_t off, int whence);
```

- `filp`: le fichier concerné par l'écriture;
- `off`: l'offset demandé (attention, pas l'offset actuel comme pour read et write);
- `whence`: l'offset doit être interprété relativement à whence (voir l'appel system `lseek`).
- retourne le nouvel offset dans le fichier.

La fonction `lseek` doit modifier `filp->f_pos` comme demandé. La fonction doit aussi collaborer avec les fonction `read` et `write`.

On peut mettre l'entrée `llseek` de la structure `f_ops` à `NULL`. Dans ce cas le noyau effectue les changements lui même (pas toujours souhaité).

--

## `lseek` - pas d'accès aléatoire

Si l'on souhaite ne pas implémenter d'accès aléatoire il faut utiliser la fonction suivante lors de l'**ouverture** du fichier:

```c
int nonseekable_open(struct inode *inode, struct file *filp);
```

Il faut aussi faire pointer l'opération vers la fonction pré-définie `no_llseek`:

```c
struct file_operations my_fops = {
    .owner = THIS_MODULE,
    ...
    .llseek = no_llseek,
    ...
};
```

--

## Autres opérations

Nous allons implémenter une version minimale d'un driver.

En fait, beaucoup d'autre opèrations doivent être implémentées pour que le drivers soit complétement opérationel.

<!-- Discuter les codes d'erreurs linux/errno.h-->
