
# Introduction

--

## Historique

<div class="cols">
<div style="font-size: 0.7em">

- Les systèmes de fichiers ont été conçu pour accéder, avec une interface simple, à des périphériques de stockage de masse (persistents)
- Les premiers périphériques de stockage de masse étaient basés sur des disques magnétiques appelés *Hard Disk Drive*
- La plus petite unité de lecture/écriture adressable est le secteur
- Adressable par le triplet CHS :\
  { Cylinder, Head, Sector }

A. Plateaux \
B. Bras mobile \
C. Têtes de lecture/écriture reliées par le bras mobile \
D. Cylindre = coupe transversale des plateaux \
E. Pistes = groupe de secteurs contigüs \
F. Secteur angulaire

</div>
<div>

![](images/hdd.png) <!-- .element width="90%" -->


</div>
</div>


--

## Historique : premier disque dur

Premier disque dur conçu en 1956 pour le super-computer IBM 305 Ramac

<div class="cols">

<div>

- Capacité : 4.8MB
  - 50 plateaux de 60cm (24"), contenant chacun 100 pistes/faces, contenant chacune 5 secteurs de 100 octets
- Débit : 8.8KB/sec.
- Vitesse : 1200 tours/min.
- 2 têtes lecture/écriture, 1 sec. pour passer d'un plateau à l'autre
- Prix du système complet : $33'200/mois

</div>
<div>

![](images/IBM-305-RAMAC.jpg)<!-- .element width="100%" -->


</div>
</div>

--

## Aujourd'hui

- Les **disques durs mécanique (HDD)** utilisent les mêmes principes (cylindres, pistes, têtes, des secteurs), mais :
  - performances et capacités ont augmenté exponentiellement
  - adressage physique CHS remplacé par adressage logique : LBA (*Logical Block Addressing*)
  - taille physique d'un secteur est souvent 4KB avec émulation à 512 bytes par le firmware
- Les **disques durs à base de  [flash](https://www.arrow.com/en/research-and-events/articles/nand-memory-explained-understanding-pros-and-cons-of-nand) (SSD)** ne comportent pas de partie mécanique, mais présentent aussi des secteurs de 512 bytes adressés logiquement en LBA (émulation réalisée par le firmware)
  - la plupart des périphiques de stockage à base de flash sont similaires (clés USB, cartes SD, etc.)


Notes:
SSD -> solid state drive
Les disque dure SSD utilisent des "floating gates transistors", transistor à grille flottante, qui matiennent l'état. On parle de cellule mémoire.

--

## Structure d'un disque

- Un disque est divisé en unités de taille identique appelées **secteurs**
- La taille d'un secteur dépend du support: **512 bytes**, 2KB, **4KB**, etc.
- Un secteur est la plus petite unité physique pouvant être lue ou écrite
- La lecture ou l'écriture d'un secteur est une opération **atomique**

![](images/sectors.png)<!-- .element width="80%" -->


--

## UNIX et blocs

- Dans un OS de type UNIX, on parle de blocs<sup>1</sup> plutôt que de secteurs
- **But : s'abstraire du type de périphérique**
- Tout périphérique dont on peut lire/écrire les données par unités de 512, 1024, etc. bytes est géré par le module noyau de
gestion de lecture/écriture par blocs
- La taille d'un bloc est potentiellement plus grande que la taille d'un secteur
- Permet d'être indépendant du type de périphérique $\rightarrow$ généricité

<small>1: Microsoft appelle un bloc un *cluster*</small>

--

## Secteurs, blocs et clusters

- **Un bloc est une collection contiguë de secteurs**
- Un système de fichiers (FS) divise l'espace disque en blocs de taille égale
- Généralement, un bloc fait entre 1KB et 8KB
- La commande `stat fichier` affiche la taille d'un bloc du FS où se trouve `fichier` (champ *IO Block*)

![](images/block_and_cluster.png)<!-- .element width="80%" -->


--

## Interface périphérique bloc

Si on fait abstraction de la gestion du cache, l'interface d'accès à un périphérique de type bloc est très simple :

- Après initialisation de la taille d'un bloc, chaque bloc est adressable logiquement via son numéro par une fonction d'écriture et une fonction de lecture :


  - `init_block_dev(dev, size)`
  - `read_blocks(dev, block_nb, buf, count)`
  - `write_blocks(dev, block_nb, buf, count)`


--

## Adressage

- Les blocs d'un périphérique de taille $N$ blocs de $n$ bytes chacun, sont adressables de `$0$` à `$N-1$` (comme un tableau en C)
- La lecture/écriture se fait uniquement par bloc, donc par unité de $n$ bytes
- La lecture/écriture du bloc $0$ traite les bytes de l'offset `$0$` à l'offset `$n-1$` du périphérique
- La lecture/écriture du bloc $b$ traite les bytes de l'offset `$n*b$` à l'offset `$n*(b+1)-1$`

![](images/addressing.png)<!-- .element width="70%" -->


--

## Adressage : exemple

Exemple : périphérique de 10MB (1024 * 1024 * 10 bytes), taille de bloc de 1024 bytes

- Quelle est l'intervalle des offsets des bytes du bloc numéro 17 ? \
  `$[1024*17,1024*(17+1)-1] = [17408,18431]$` <!-- .element: class="fragment" data-fragment-index="0" -->
- Dans quel bloc se trouve le byte localisé à l'offset 7000 du périphérique ? \
  `$\lfloor{7000/1024}\rfloor = \lfloor{6.8359375}\rfloor = 6$` <!-- .element: class="fragment" data-fragment-index="1" -->
- À quel offset du bloc est localisé le byte se trouvant à l'offset 7000 du périphérique ? \
  `$7000 - 6 * 1024 = 856 = 7000 \% 1024$` <!-- .element: class="fragment" data-fragment-index="2" -->

![](images/addressing_example.png)<!-- .element width="70%" -->



--

## Abstraction (1/2)

- Avec cet adressage logique, on peut facilement émuler un périphérique bloc à partir d'un fichier image
- On peut y accéder via les appels systèmes usuels `read` et `write`
- Ces appels système peuvent simplement être mappés sur les fonctions décrites précédemment :

<div style="font-size: 0.9em">

  - `init_block_dev(dev, size)` $\longrightarrow$ `fd = open(file)`
  - `read_blocks(dev, block_nb, buf, count)` $\longrightarrow$ `read(fd)`
  - `write_blocks(dev, block_nb, buf, count)` $\longrightarrow$ `write(fd)`

</div>

--

## Abstraction (2/2)

- Le fichier image peut-être distant, si on implémente l'abstraction au dessus des sockets et d'un protocole qui
reste très simple :

  - `init_block_dev(dev, size)` $\longrightarrow$ `s = socket()`
  - `read_blocks(block_nb, buf, count)` $\longrightarrow$ `read(s)`
  - `write_blocks(block_nb, buf, count)` $\longrightarrow$ `write(s)`


--

## Conclusion

- On peut lire/écrire des blocs adressés logiquement, mais on est encore loin d'un FS <!-- .element: class="fragment" data-fragment-index="0" -->
- Comment à partir de ces opérations simples, peut-on construire un FS ? <!-- .element: class="fragment" data-fragment-index="2" -->
- Pour cela, il est nécessaire de comprendre l'allocation des blocs de données ainsi que l'organisation sur disque d'un FS <!-- .element: class="fragment" data-fragment-index="3" -->
