<!--
## TODO

- Update FAT diagram:
  - indicate a file's metadata with: first block, file size (along the type), type, etc.
- Update inode diagram:
  - indicate a file's inode with: file size (along the type), permission, type, etc. and block pointers
  - specify an inode has a fixed size
  - update diagram to show the data blocks to "scale", especially vs an inode
- ext4: indicate that very small files' content is stored in the inode and don't use a data block
- would be good to the number of single ptr, indirect ptr, etc. for ext2, ext3 and ext4
-->

# Stratégies d'allocation

--

## Système de fichiers - rappel

- Un système de fichiers (FS) est un ensemble de fichiers
- Les fichiers stockent des données
- Les fichiers et les répertoires sont tous deux des fichiers, mais de types différents
- Du point de vue du FS :
  - un fichier ordinaire ou régulier contient... les données du fichier
  - un répertoires contient des entrées de répertoire
- Fichiers et répertoires sont représentés de la même manière, mais leurs contenus sont différents

--

## Fichier - rappel

Un fichier est composé de deux parties :

- **Métadonnées**
  - type, permissions, (nom), propriétaire, pointeurs vers les données, etc.
- **Données (contenu)**
  - fichier "ordinaire" : blocs contenant le contenu du fichier
  - répertoire : blocs contenant les références vers les fichiers présents dans le répertoire

--

## Défi

- Comment organiser un FS hiérarchique en utilisant des blocs comme unité de base ?
- Comment rendre le FS performant ?
  - en termes d'espace (pas d'espace perdu)
  - en terme de lecture et d'écriture de fichiers

--

## Stratégies d'allocation des blocs de données

- La création d'un fichier ou d'un répertoire nécessite **d'allouer des blocs**
- La stratégie d'allocation des blocs de données (contenu) doit :
  - être efficace en termes d'espace
  - garantir un accès rapide aux fichiers

--

## Métriques d'allocation de blocs

Métriques principales:

- Les fichiers peuvent-ils grandir ?
- Performance des accès séquentiels
- Performance des accès aléatoires
- Facilité d'implémentation
- Efficacité du stockage/*overhead* (espace utilisé pour les mécanismes de stockage vs. données effectivement stockées)
- Fragmentation

--

## Fragmentation interne

<div class="cols">
<div>

- L'espace alloué peut être plus grand que l'espace demandé
- **L'espace alloué non utilisé est gaspillé !**

</div>
<div>

![](images/internal_fragmentation.png)<!-- .element width="80%" -->

</div>
</div>

--

## Fragmentation externe

<div class="cols">
<div>

- Bien qu'il y ait suffisamment d'espace libre au total, il n'y a pas assez d'espace libre contigu pour satisfaire la demande d'allocation
- **Impossible de satisfaire la demande d'allocation !**

</div>

<div>

![](images/external_fragmentation.png)

</div>
</div>

--

## Principales stratégies d'allocation de blocs

- Le choix de la stratégie d'allocation dépend :
  - de la technologie de stockage (SSD, HDD, etc.)
  - de la stratégie d'accès et le type d'utilisation
    - lecture seule, gros fichiers, petits fichiers, fichiers grandissant beaucoup, etc.

- Stratégies d'allocation principales
  1. contiguë
  1. liste chaînée
  1. indexée

--

## 1. Allocation contiguë

- Chaque fichier utilise un ensemble de blocs contigus
- Un fichier est représenté par:
  - l'index du premier bloc du fichier
  - la taille du fichier

--

## 1. Allocation contiguë : exemple

![](images/contiguous_alloc.png)<!-- .element width="60%" -->

--

## 1. Allocation contiguë : synthèse

<fieldset class="OK">
<legend>Avantages</legend>

- très facile à implémenter
- accès séquentiel rapide
- accès aléatoire rapide
- faible *overhead* stockage (structure de données très simple)

</fieldset>
<p></p>
<fieldset class="warning fragment">
<legend>Inconvénients</legend>

- les fichiers ne peuvent pas (ou peu) grandir
- fragmentation externe importante
- fragmentation interne lorsque taille fichier < taille bloc

</fieldset>

--

## 2. Allocation par liste chaînée

- Un fichier est représenté par:
  - un pointeur (index) sur le premier bloc
  - la taille du fichier (ou un pointeur sur le dernier bloc)

- Les bloc sont stockés sous forme de liste chainée:
  - chaque bloc possède un pointeur (index) sur le bloc suivant

  ![](images/linked_list.png)

--

## 2. Allocation par liste chaînée : exemple

![](images/linked_list_alloc.png)<!-- .element width="55%" -->

--

## 2. Allocation par liste chaînée : synthèse


<fieldset class="OK">
<legend>Avantages</legend>

- les fichiers peuvent facilement grandir, sans limite
- facile à implémenter, mais le stockage des pointeurs est délicat
- pas de fragmentation externe

</fieldset>
<p></p>
<fieldset class="warning fragment">
<legend>Inconvénients</legend>

- accès séquentiel lent
- accès aléatoire lent : adresse difficile à calculer
- fragmentation interne lorsque taille fichier < taille bloc

</fieldset>

--

## 2.b. Allocation par FAT

<div class="cols">
<div style="width:200%">

- Variante de l'allocation par liste chaînée
- Les pointeurs de blocs sont stockés dans une table dédiée située au début du FS
  - taille de la table d'allocation des fichiers (FAT)
  - <small>une entrée FAT par bloc de données</small>
  - <small> liste chaînée d'entrées FAT par fichier</small>
  - <small> valeur spéciale indique la fin de la liste (EOC = *End of Chain*)</small>
- Système fichiers de MS-DOS et Winows 95, utilisé dans cartes SD et clés USB
- Nombreuses variantes : FAT16, FAT32, exFAT, etc.

</div>
<div>

![Structure sur disque du FS FAT](images/FAT_disk_layout.png)<!-- .element width="90%" -->

<small>Structure sur disque du FS FAT</small>

</div>
</div>

--

## 2.b. Contenu de la FAT : exemple

![](images/FAT_alloc.png)

--

## 2.b. FS de type FAT : exemple

<div style="font-size:0.8em">

Soit le FS de type FAT suivant : <!-- .element: style="text-align:left;margin:0px" -->

- Le FS comporte 1'000 blocs de données
- Taille des blocs de données : 4 KB
- Chaque entrée dans la FAT est codée sur 16 bits (une valeur est réservée pour représenter la EOC - *End Of Chain*)

</div>

**Questions**

1. Quelle est la taille de la FAT en bytes pour ce FS ? \
 `$1000*(16 bits) = 2000$ bytes` <!-- .element: class="fragment" data-fragment-index="1" -->

1. Quelle est la taille de fichier maximum supportée pour ce FS (en bytes, KB et MB) ? \
`$1000*4096 = 4096000$ bytes, $4000$ KB, $3.9$ MB` <!-- .element: class="fragment" data-fragment-index="2" -->

1. Combien de blocs de données cette FAT pourrait-elle gérer au maximum théoriquement ? \
`$(2^{16})-1 = 65535$ blocs` <!-- .element: class="fragment" data-fragment-index="3" -->


--

## 2.b. Allocation par FAT : synthèse

<fieldset class="OK">
<legend>Avantages</legend>

- relativement facile à implémenter
- les fichiers peuvent facilement grandir (tant qu'il y a de l'espace dans la FAT)
- accès aléatoire rapide
- pas de fragmentation externe

</fieldset>
<p></p>
<fieldset class="warning fragment">
<legend>Inconvénients</legend>

- accès séquentiel lent si les blocs ne sont pas contigus (HDD uniquement)
- *overhead* important pour le stockage de la FAT (disque et RAM), en particulier avec un grand nombre de blocs/clusters
- la FAT doit être chargée en RAM (sinon performances catastrophiques)
- fragmentation interne lorsque taille fichier < taille bloc

</fieldset>

--

## 3. Allocation indexée

- Chaque fichier est associé à un **inode** de taille fixe
- L'inode contient
  - les métadonnées du fichier
  - la liste des pointeurs vers les blocs de données

![](images/indexed_alloc.png) <!-- .element width="45%" -->

--

## 3.b. Allocation indexée multi-niveau

- L'inode stocke pointeurs directs, indirects, doublement indirects, etc.
- Utilisé par tous les FS dans les systèmes UNIX : minix-fs, ext2, ext3, etc.

![](images/multi_indexed_alloc.png) <!-- .element width="85%" -->

--

## 3.b. Alloc. indexée multi-niv. : exemple

<div style="font-size:0.8em">

Soit le FS de type indexé multi-niveau suivant : <!-- .element: style="text-align:left;margin:0px" -->

- Un inode fait 64 bytes
- Un inode contient 8 pointeurs directs, 2 pointeurs indirects et 1 pointeur doublement indirect
- Un pointeur de bloc est stocké sur 16 bits
- Taille des blocs de données : 1 KB (1024 bytes)

</div>

**Questions**

1. Combien peut-on stocker de pointeurs par bloc ? \
   `$1024/2 = 512$` <!-- .element: class="fragment" -->

2. Quelle est la taille de fichier maximum supportée pour ce FS (en bytes, KB et MB) ? \
  `$8*1024 + (512*1024)*2 + (512^2)*1024 = 269492224$ bytes, $263176$ KB, $257$ MB` <!-- .element: class="fragment" -->

<!--
1. Combien peut-on stocker d'inodes par bloc\ ?
   - \footnotesize \textcolor{mygreen}{$1024/64$ = 16}
-->

--

## 3.b. Allocation indexée : synthèse

<fieldset class="OK">
<legend>Avantages</legend>

- relativement facile à implémenter
- les fichiers peuvent facilement grandir (tant qu'il y a des pointeurs libres)
- accès aléatoire rapide
- pas de fragmentation externe

</fieldset>
<p></p>
<fieldset class="warning fragment">
<legend>Inconvénients</legend>

- *overhead* de stockage pour les pointeurs
- l'accès rapide nécessite l'allocation de blocs contigus (HDD uniquement)
  - sinon, accès potentiellement lent
- fragmentation interne lorsque taille fichier < taille bloc

</fieldset>

--

## 3.c. Allocation par extent

- Principe similaire à l'allocation indexée
- **Différence** : un pointeur référence un *extent* plutôt qu'un bloc
- *Extent* = ensemble de blocs contigüs
  - représenté par le tuple : `{ FirstBlockAddress, Length }`
- Exemple : 64 bits par *extent* :
  - 48 bits pour l'indice du 1er bloc
  - 16 bits pour la longueur (= nombre de blocs)
- **Avantages** : moins de blocs à stocker si la plupart des allocations sont contiguës
- Utilisé par les FS "modernes" : ext4, btrfs, ntfs, xfs, etc.

--

## 3.c. Allocation par extent : synthèse

<fieldset class="OK">
<legend>Avantages</legend>

- les fichiers peuvent grandir (tant qu'il y a des extents libres)
- accès séquentiel rapide
- accès aléatoire rapide
- faible *overhead* stockage (structure de données simple)

</fieldset>
<p></p>
<fieldset class="warning fragment">
<legend>Inconvénients</legend>

- fragmentation externe potentielle
- plus complexe que l'allocation indexée
- fragmentation interne lorsque taille fichier < taille bloc
</fieldset>
